<?php
/*
	TODO: 
	- 8-5a, 8-5b input pipes.
	DONE: 
	+ Map doodads.
	+ Reactor numbering.
	+ Check reactor positioning for CV-6, CV-7, 9-1.
	+ Extra pipe for non-extended inputs.
	+ Building clipping (i.e., 4-C's huge_powerplant)
	+ Elimination of double buildings.
	+ Basic building placement (input, output, reactors, special)
	+ Pipelining.
	+ Map tiling.
	+ Map BG.
	
*/


class SpaceChemMap
{
	// ------------------------------------------------------------------------
	// CONSTANTS
	// ------------------------------------------------------------------------

	const TILE_WIDTH= 24;
	const TILE_HEIGHT= 21;
	const MAP_COLS= 32;
	const MAP_ROWS= 23;		// 1 dummy that's always empty.
	
	// ------------------------------------------------------------------------
	// MEMBERS
	// ------------------------------------------------------------------------

	private $_key;			//!< Map key.
	private $_ldb;			//!< SQLite3 save file.
	private $_map;			//!< Overworld map data.
	private $_bgCss;		//!< BG CSS style
	private $_buildings;	//!< Building information.
	
	private $_tiles;		//!< Tile data for final rendering.
	
	// ------------------------------------------------------------------------
	// PROPERTIES
	// ------------------------------------------------------------------------

	public function key()			{	return $this->_key;				}
	public function map()			{	return $this->_map;				}
	public function buildings()		{	return $this->_buildings;		}
	
	// ------------------------------------------------------------------------
	// METHODS
	// ------------------------------------------------------------------------
	
	public function __construct()
	{
		$this->_ldb= null;
		
		$this->init();
	}
	
	public function init()
	{
		$tiles= array_fill(0, SpaceChemMap::MAP_ROWS, null);
		for($iy=0; $iy<SpaceChemMap::MAP_ROWS; $iy++)
		{
			$tiles[$iy]= array_fill(0, SpaceChemMap::MAP_COLS, array(
				'border'	=> 0,
				'buildings'	=> array(),
				'pipes'		=> array(),
			));
		}
		$this->_tiles= $tiles;
		
		$this->_key= "";
		$this->_map= array();
		$this->_bgCss= "";
		$this->_buildings= array();
	}
	
	public function load($ldb, $key)
	{
		$this->_ldb= $ldb;
		
		$res= mysql_query("SELECT * FROM `maps` WHERE `key` = '$key'");
		
		$map= mysql_fetch_assoc($res);
		if(~((int)($map['type'])) & 1)
		{
			echo "<p>Not a production map</p>";
			return false;
		}
		
		$this->_key= $key;
		$this->_map= $map;
		
		$this->_bgCss= "background: url(./img/bg/{$map['bgImage']}) no-repeat 0 -{$map['bgY']}px";
		
		$this->loadTiles();
		$this->loadBuildings();
		$this->loadPipes();
		
		// Extras for testing:
		$buildings= &$this->_buildings;
		$tiles= &$this->_tiles;
		
		return true;
	}
	
	public function render()
	{
		$tiles= $this->_tiles;
		$map= $this->_map;
		$buildings= $this->_buildings;
		
		// --- Pre-render buildings ---
		$rxCount= 0;
		foreach($buildings as $b)
		{
			// --- Reactor numbering ---
			$index= "";
			if($b['type'] == "r")
			{
				$rxCount++;
				$index= "<a class=\"rx\" href=\"#$rxCount\">$rxCount</a>";
			}
			
			
			// --- Clipping ---
			// (That means YOU, huge_powerplant >:[ )
			$x= $b['x'];
			$y= $b['y'];
			
			$margin= "";
			
			if($x<0)
			{
				$margin .= "margin-left: ". $x*SpaceChemMap::TILE_WIDTH ."px; "; 
				$x= 0;
			}
			if($y<0)
			{
				$margin .= "margin-top: ". $x*SpaceChemMap::TILE_HEIGHT ."px; "; 
				$y= 0;
			}
			$style= (empty($margin)) ? "" : " style=\"$margin\"";
			
			// --- Pre-render output ---
			$tiles[$y][$x]['buildings'][]= "<li class=\"{$b['class']}\" $style>$index</li>\n";
		}
		
		
		// --- Final render ---
		
		$content= "";
		for($iy=0; $iy<SpaceChemMap::MAP_ROWS; $iy++)
		{
			$row= "";
			for($ix=0; $ix<SpaceChemMap::MAP_COLS; $ix++)
			{
				$tile= $tiles[$iy][$ix];
				$class= ($tile['border']!= "" ? " class=\"{$tile['border']}\"" : "");
				
				$parts= implode($tile['buildings']);
				foreach($tile['pipes'] as $v)
					$parts .= "<li class=\"pipe-$v\"></li>\n";
				
				if($parts != "")
					$parts= "<ul>$parts</ul>";
				$row .= "   <td$class>{$parts}</td>\n";
			}
			$content .= "  <tr>\n$row  </tr>\n";
		}
		
		$content= <<<DOC
<table class="map" cellpadding=0 cellspacing=0 style="{$this->_bgCss}">
$content
</table>
DOC;
		return $content;
	}
	
	
	private function loadTiles()
	{
		$map= $this->_map;
		$tiles= &$this->_tiles;
		
		// --- Init tile borders ---
		// Each row in fields is a 32bit number with bitflags indicating buildable
		// tiles. The following compares the current tile to the ones on the right and 
		// below it, and issues the right type of border. Default border is bottom+right.
		
		$fields= preg_split("#\s+#", $map['fields']);
		for($iy=0; $iy<SpaceChemMap::MAP_ROWS; $iy++)
			$fields[$iy]= hexdec(aget($iy, $fields, 0));
		$fields[SpaceChemMap::MAP_ROWS]= 0;
		
		$borders= array("bdr-0", "bdr-r", "bdr-b", "");
		for($iy=0; $iy<SpaceChemMap::MAP_ROWS; $iy++)
		{
			for($ix=0; $ix<SpaceChemMap::MAP_COLS; $ix++)
				$tiles[$iy][$ix]['border']= $borders[
					( ($fields[$iy]|$fields[$iy]<<1)>>(31-$ix)&1 ) | 
					( ($fields[$iy]|$fields[$iy+1] )>>(31-$ix)&1 )*2
				];
		}
	}
	
	private function loadBuildings()
	{
		$key= $this->_key;
		$ldb= $this->_ldb;
		$map= $this->_map;
		$tiles= &$this->_tiles;
		
		// --- Get savegame buildings ---
		$buildings= ldbQueryAll($ldb, "SELECT c.* FROM `Component` c 
			JOIN `Level` lvl ON c.level_id = lvl.id 
			WHERE lvl.id = '{$map['level_id']}' ORDER BY c.y, c.x");
		
		$rxIndex= 1;
		foreach($buildings as &$building)
		{
			// Special-case for 9-1's superlaser.
			if($key == "9-1" && $building['type'] == "drag-superlaser-reactor")
					$building['type']= "drag-hyperlaser-reactor";
					
			// Combine with DB's building details.
			$res= mysql_query("SELECT * FROM `buildings` WHERE `component_id`='{$building['type']}'");
			if($res && mysql_num_rows($res))
				$building= array_merge($building, mysql_fetch_assoc($res));
		}
		
		// --- Get map buildings from DB ---
		$res=  mysql_query("SELECT * FROM `map_buildings` mb 
			JOIN `buildings` b ON mb.component_id = b.component_id 
			WHERE `level_id`='{$map['level_id']}'");

		$count= count($buildings);
		while($row= mysql_fetch_assoc($res))
		{
			// Eliminate duplicates, based on type and coordinates.
			for($i=0; $i<$count; $i++)
			{
				$b= $buildings[$i];
				if($row['component_id']==$b['component_id'] && $row['x']==$b['x'] && $row['y'] == $b['y'])
					break;
			}
			
			$row['rowid'] = -1;
			if($count==0 || $i >= $count)
				$buildings[]= $row;
		}
	
		$this->_buildings= $buildings;
	}
	
	private function loadPipes()
	{
		$ldb= $this->_ldb;
		$buildings= $this->_buildings;
		$tiles= &$this->_tiles;

		// Look-up tables. This will tie adjacent pipes to the proper CSS classes.
		$lutX= array(0, 0, 1);
		$lutY= array(2, 0, 3);
		$lutClass= array(
			array( "x", "lr", "lu", "ld"), 				// From left
			array("lr",  "x", "ru", "rd"), 				// From right
			array("lu", "ru",  "x", "ud"), 				// From up
			array("ld", "rd", "ud",  "x"), 				// From down
		);
	
		foreach($buildings as $building)
		{
			$id= $building['rowid'];
			$outCount= $ldb->querySingle("SELECT COUNT(DISTINCT `output_id`) FROM `Pipe` 
				WHERE component_id='$id'");
				
			$x0= $building['x'];
			$y0= $building['y'];
			
			if($outCount > 0)
			{
				// Building has pipes.
				for($i=0; $i<$outCount; $i++)
				{
					$a= ldbQueryAll($ldb, "SELECT x,y FROM `Pipe` 
						WHERE component_id='$id' AND output_id='$i' ORDER BY rowid;");
					
					// Add extra dummy-piece to curve into next building.
					// #TODO: Replace with proper check later.
					$count= count($a);
					$a[]= array('x' => $a[$count-1]['x']+1, 'y' => $a[$count-1]['y']);
					
					$curr= 0;					// Start with entry from left.
					for($j=0; $j<$count; $j++)
					{
						$x= $a[$j]['x'];
						$y= $a[$j]['y'];
						$dx= $a[$j+1]['x'] - $x;
						$dy= $a[$j+1]['y'] - $y;
						
						$next= $lutX[$dx+1] + $lutY[$dy+1];
						$tiles[$y0+$y][$x0+$x]['pipes'][]= $lutClass[$curr][$next];
						
						$curr= $next^1;			// Reverse direction.
					}
				}
			}
			else if($building['type'] == "i")
			{
				// If there is no out-count, and it's an input, 
				// create a dummy pipe. NOTE: not perfected yet.
				$tiles[$y0][$x0+$building['width']]['pipes'][]= "lr";
			}
		}
	}
}

?>